#!/usr/bin/perl
use warnings;
use strict;
use FindBin qw($Bin);
use lib "$Bin/modules";

use Cwd;
use Math::Trig;
use get_info;

# setup_rmc.pl is used to setup the random model calculation (RMC). It
# does not actually launch any of the parallel image processing codes,
# but rather generates the directory structure, auto3dem parameter
# files, random particle orientations, and command script required to
# carry out the RMC

# Input: All input optional and specified by flags. Defaults values
# shown in [ ] following descriptions
#
# -boxrad=n       image box size [obtained from image file]
# -list=name      name of file listing particle parameter files [list]
# -ncpu=n         number of CPUs [4]
# -nimages=n      number of images to use in constructing model [150]
# -nmodels=n      number of random models [10]
# -noctf          turn off CTF corrections
# -nodefile=name  name of file containing list of compute nodes
# -rundir=name    directory containing particle parameter files [dat]
# -seed=n         seed for random number generator. Use an integer for
#                 reproducible result, -seed=time for seed based on
#                 seconds since 1/1/1970 [1]
#
# ----- Directory structure for RMC -----
#
# The user specifies the directory containing the particle image and
# parameter files. Within this directory, subdirectories are then
# created for each random model: RMC1, RMC2, etc. Within the
# subdirectories new parameter files are written which contain the
# random particle orientations.
#
# ----- Random orientation generation -----
#
# First generate list of (theta,phi) values evenly covering 1/2 of the
# asymmetric unit (ASU), typically done using spacing of 0.5
# degrees. Then, for each particle, randomly choose
#
# (1) (theta,|phi|) from set of orientations
# (2) Sign of phi
# (3) Omega in the range 0 to 360
# 
# The new particle parameter files contain the same header information
# as the original files, but with random orientations and particle
# centers initialized to lie at the center of the box.

# ----- auto3dem parameter files -----
#
# For each random model, an auto3dem parameter file is
# created. Default values are heavily relied upon and only a minimal
# set of parameters are explicitly set. These include the name of the
# final map, the run directory, resolutions, and annulus sizes. The
# initial choices for a number of these parameters depend explicitly
# on the box and pixel sizes of the images.

# ----- Command file -----
#
# The RMC can be broken down into a set of independent auto3dem runs
# each using a different set of random particle orientations. This
# script generates a command file that launches the individual
# auto3dem runs and parses the results to determine which random model
# results in the best resolution.

my $version_string = "------ SETUP_RMC version 3.06 ------";

# Save the starting directory
my $startdir = getcwd();

# Get input from command line and use to overwrite default values

# Print usage info if no arguments provided
if(@ARGV == 0 || $ARGV[0] =~ /^-help/i) {
print << "EOF";
$version_string

NAME
  setup_rmc - script for setting up random model calculation 

EXAMPLE 
  setup_rmc -ncpu 4 -seed 123 -list listfile
  setup_rmc -usedefaults

DESCRIPTION
  All input specified using the syntax -key value. Defaults values are
  shown in brackets following descriptions. To use all default values,
  use the -usedefaults flag. 'n' = integer value, 's' = string

  -boxrad n    image box size [obtained from image file]
  -list s      file or expression listing particle parameter files 
               If not specified, then setup_rmc looks for 
               (1) file named 'list' or (2) *00n files in rundir
  -ncpu n      number of CPUs [4]
  -nimages n   number of images to use in constructing model [150]
  -nmodels n   number of random models [10]
  -noctf       turn off CTF corrections
  -nodefile s  file containing list of compute nodes
  -rundir s    directory containing particle parameter files [dat]
  -seed n      seed for random number generator. Use an integer for
               reproducible result or -seed=time for seed based on
               seconds since 1/1/1970 [1]

NOTES 
  Nodefile specification depends on your computing environment and in
  many cases it can be omitted. For systems running the PBS scheduler,
  the nodefile is automatically obtained from the environment assuming
  that "#PBS -V" was used in the batch script. Keys are case
  insensitive (e.g. -ncpu, -Ncpu, -NCPU all OK), but string values are
  case sensitive. You will rarely need to set boxrad since it is
  normally determined from image file.

*** Continue running setup_rmc using all default values? ***
Enter [y]es to continue
EOF

    my $continue = <>;
    unless($continue =~ /y/i || $continue =~ /yes/i) {
	exit(0);
    }
}

print "--- setup_rmc starting ---\n\n";

if(defined $ARGV[0] && 
   ($ARGV[0] =~ /^-usedefaults$/i || $ARGV[0] =~/^-help$/i)  ) {
    @ARGV = ();
}

# Process command line args so that we can split on ' -'
my $command_line = join ' ', @ARGV;
$command_line =~ s/^\s+//;   # Get rid of leading whitespace
$command_line =~ s/\s+$//;   # Get rid of trailing whitespace
$command_line =~ s/=/ /g;    # Change '=' to ' '
$command_line =~ s/^-//;     # Get rid of leading '-'
@ARGV = split(/\s+-/, $command_line);

my %valid = (boxrad => 1, list => 1, ncpu => 1, nimages => 1,
	     nmodels => 1, noctf => 1, nodefile => 1, rundir => 1,
	     seed => 1, help => 1, usedefaults => 1);

my %cmdline_arg;
foreach my $arg (@ARGV) {
    my($key, $value) = split(/\s+/, $arg);
    $key =~ tr/A-Z/a-z/;
    unless($valid{$key}) {
	print "\n--- Error ---\n";
	print "  -$key is invalid keyword. Must be one of the following\n\n";
	foreach my $key (sort keys %valid) {
	    print "  -$key\n";
	}
	print "\nType setup_rmc without any arguments to see usage info\n";
	exit(0);
    }
    $cmdline_arg{$key} = $value;
}

my ($boxrad_cmdline, $list, $ncpu, $nimages, $nmodels, $nodefile,
    $rundir, $seed);

print "setup_rmc parameters:\n";

if(defined $cmdline_arg{boxrad}) {
    $boxrad_cmdline = $cmdline_arg{boxrad};
    print "  User specified boxrad = $boxrad_cmdline\n";
}

if(defined $cmdline_arg{rundir}) {
    $rundir = $cmdline_arg{rundir};
    print "  User specified rundir = $rundir\n";
} else {
    $rundir = 'dat';
    print "  Default rundir = $rundir\n";
}

if(defined $cmdline_arg{list}) {
    $list = $cmdline_arg{list};
    print "  User specified list = $list\n";
} else {
    print "  list not specified - setup_rmc will first check for\n";
    print "    file 'list', then look for files $rundir/*00n\n";
}

if(defined $cmdline_arg{ncpu}) {
    $ncpu = $cmdline_arg{ncpu};
    print "  User specified ncpu = $ncpu\n";
} else {
    $ncpu = 4;
    print "  Default ncpu = $ncpu\n";
}

if(defined $cmdline_arg{nimages}) {
    $nimages = $cmdline_arg{nimages};
    print "  User specified nimages = $nimages\n";
} else {
    $nimages = 150;
    print "  Default nimages = $nimages\n";
}

if(defined $cmdline_arg{nmodels}) {
    $nmodels = $cmdline_arg{nmodels};
    print "  User specified nmodels = $nmodels\n";
} else {
    $nmodels = 10;
    print "  Default nmodels = $nmodels\n";
}

if(defined $cmdline_arg{nodefile}) {
    $nodefile = $cmdline_arg{nodefile};
    print "  User specified nodefile = $nodefile\n";
    unless($nodefile =~ /^\//) {
	$nodefile = "$startdir/$nodefile";
    }
}

if(defined $cmdline_arg{seed}) {
    $seed = $cmdline_arg{seed};
    if($seed =~ /time/i) {
	$seed = time();
	print "  User specified seed based on time = $seed\n";
    } elsif ($seed =~ /^\d+$/) {
	print "  User specified seed = $seed\n";
    } else {
	print "\n--- Error ---\n";
	print "  Invalid seed specified (seed = $seed)\n";
	print "  Seed must be integer or the string 'time'\n";
	exit(0);
    }
} else {
    $seed = 1;
    print "  Default seed = $seed\n";
}

# Initialize random number generator
srand($seed);

# Fill arrays of views (theta, phi) in the ASU
my $dang = 0.5;
my @theta = ();
my @phi = ();
get_all_views($dang, \@theta, \@phi);
my $nviews = @theta;

my %ptlefile;
my %head;
my %numimages;
my %boxrad;
my %pixel_size;
my %defocus;
my ($defocus_maj, $defocus_min);

# Get particle parameter files from $list. Note the $list can be
# either a file or an expression. If $list was not defined, then try
# defaults (1) look for file named 'list' or (2) look for *00n files
# in the $rundir directory. Expand wildcards as necessary.

my @list = ();
if(defined $list) {
    if(-r $list) { # Option 1 - $list is a readable file
	open(LIST, "$list");
	while(<LIST>) {
	    my $record = $_;
	    $record =~ s/#.*$//g; # Get rid of comments
	    unless(defined $record && $record =~ /\w/) {
		next;
	    }
	    $record =~ s/\s+//g;
	    push(@list, glob("${rundir}/$record"));
	}
	close(LIST);
    } else { # Option 2 - $list is an expression
	$list =~ s/^.*\///; # Strip off path information
	$list =~ s/\s+//g;  # Get rid of whitespace
	@list = glob("${rundir}/$list");
    }
} else {
    if(-r 'list') { # Option 3 - $list not specified, try to open 'list'
	open(LIST, 'list');
	while(<LIST>) {
	    my $record = $_;
	    $record =~ s/#.*$//g; # Get rid of comments
	    unless(defined $record && $record =~ /\w/) {
		next;
	    }
	    $record =~ s/\s+//g;
	    push(@list, glob("${rundir}/$record"));
	}
	close(LIST);
    } else { # Option 4 - $list not specified, look for files in $rundir
	for (my $i=0; $i<10; $i++) {
	    my $suffix = "00${i}";
	    @list = glob("${rundir}/*${suffix}");
	    if(@list > 0) {
		last;
	    }
	}
    }
}
    
# Get rid of path information and whitespace in particle parameter files
foreach my $ptlefile (@list) {
    $ptlefile =~ s/.*\///;
    $ptlefile =~ s/\s+//g;
}

# Make sure that we have at least one file in @list
unless(@list > 0) {
    print "\n--- Error ---\n";
    print "  No particle parameter files found\n";
    if(defined $list) {
	print "  $list is empty or doesn't expand to at least one file\n";
    } else {
	print "  File 'list' does not exist and cannot find files\n";
	print "  $rundir/*000, $rundir/*001, etc.\n";
    }
    exit(0);
}

# Make sure that all files in @list are readable
foreach my $ptlefile (@list) {
    unless(-r "${rundir}/${ptlefile}") {
	print "\n--- Error ---\n";
	print "  File ${rundir}/${ptlefile} is unreadable\n";
	exit(0);
    }
}

print "\nParameter files in $rundir considered for model construction:\n";
foreach my $ptlefile (@list) {
    print "  $ptlefile\n";
}
print "\n";

# For each particle parameter file, get the name of the corresponding
# PIF image file, header information, number of images, box radius,
# and pixel size.

chdir($rundir);
foreach my $ptlefile (@list) {
    open(PTLE, $ptlefile);
    my $pif = <PTLE>;
    $pif =~ s/\s+//g;
    $head{$pif} = <PTLE>;
    chomp($head{$pif});
    (undef, undef, undef, undef, $defocus_maj, $defocus_min, 
     undef, undef) = split(/,/, $head{$pif});
    $defocus{$pif} = ($defocus_maj + $defocus_min)/2.0;
    close(PTLE);
    
    $numimages{$pif} = get_numimages($pif);
    $boxrad{$pif} = get_boxrad($pif);
    $pixel_size{$pif} = get_pixel_size($ptlefile);
    $ptlefile{$pif} = $ptlefile;
}
chdir($startdir);

# Use a total of $nimages particles to construct the random
# model. Sort the micrographs by defocus and select particles starting
# with the highest defocus micrographs until the limit is reached. The
# array @pifs will contain the list of micrographs to be used and the
# corresponding values of the hash %numimages are reset to the
# required number of images from each micrograph.

my @pifs = ();
my $running_total = 0;
foreach my $pif (sort { $defocus{$b} <=> $defocus{$a} } keys %defocus) {
    push(@pifs, $pif);
    if($running_total + $numimages{$pif} >= $nimages) {
	$numimages{$pif} = $nimages - $running_total;
	last;
    }
    $running_total += $numimages{$pif};
}


# Use first file to get boxrad and pixelsize
my $boxrad_from_pif = $boxrad{$pifs[0]};
my $pixel_size = $pixel_size{$pifs[0]};
my $boxrad;

if($boxrad_from_pif == 0) {
    undef $boxrad_from_pif;
}

# The following code is used to allow for the specification of the
# box radius in either the PIF file or the auto3dem parameter
# file. Necessary because some PIF files are missing the box
# radius

# Case 1 - box radius not set in either location
unless(defined $boxrad_cmdline || defined $boxrad_from_pif) {
    print "Box radius not specified in PIF file or command line\n";
    print "Either fix $pifs[0] or manually specify box radius\n";
    die "Terminating execution\n";
}

# Case 2 - box radius specified in both locations
if(defined $boxrad_cmdline && defined $boxrad_from_pif) {
    $boxrad = $boxrad_cmdline;
    print "Box radius specified in both PIF file and command line\n";
    if($boxrad_cmdline == $boxrad_from_pif) {
	print "Box radius ($boxrad) specified consistently\n";
    } else {
	print "\nWarning boxrad from PIF file ($boxrad_from_pif) ";
	print "not equal to boxrad from command line ";
	print "($boxrad_cmdline)\n";
	print "Value in parameter file overides PIF file\n";
    }
}

# Case 3 - box radius specified only in PIF file
if(defined $boxrad_from_pif && !defined $boxrad_cmdline) {
    $boxrad = $boxrad_from_pif;
    print "Box radius from PIF file = $boxrad\n";
}

# Case 4 - box radius specified only in command line
if(!defined $boxrad_from_pif && defined $boxrad_cmdline) {
    $boxrad = $boxrad_cmdline;
    print "Box radius from command_line = $boxrad\n";
}


# Create directory structure, set symbolic links, generate files
# containing random particle orientations, write auto3dem parameter
# files, and create command file for running RMC

my $logfile_list = "RMC_logfile_list";
open(LOGLIST, ">$logfile_list");

open(CMD, ">RMC_run");

print CMD "#!/bin/sh\n\n";
print CMD "rm -f RMC_DONE\n";
print CMD "for (( i = 1 ; i <= $nmodels ; i++ ))\n";
print CMD "do\n";
print CMD "  if [ -f RMC_DONE ] ; then\n";
print CMD "    rm -f RMC_DONE\n";
print CMD "    break\n";
print CMD "  fi  \n";
print CMD "  auto3dem -ncpu $ncpu -input RMC\$[i]_master";
if(defined $nodefile) {
    print CMD "-nodefile=$nodefile\n";
} else {
    print CMD "\n";
}
print CMD "done";
print CMD "\n";

for (my $i=1; $i<=$nmodels; $i++) {
    print LOGLIST "RMC${i}_log\n";

    open(AUTO, ">RMC${i}_master");

    print AUTO "# Parameter file for random model $i\n\n";
    print AUTO "auto rundir          ${rundir}/RMC${i}\n";
    print AUTO "auto mode            search\n";
    print AUTO "auto switch_mode     0 # Don't allow switch to refine mode\n";
    print AUTO "auto niter           10\n";
    print AUTO "auto have_map        0\n";
    print AUTO "auto outfile         RMC${i}\n";
    print AUTO "auto freeze_annulus  1\n";
    print AUTO "auto freeze_res      1 # Use only for RMC\n";
    print AUTO "auto boxrad          $boxrad\n";
    print AUTO "auto quit_early      1 # Quit if fcs_hithresh not reached\n";
    if(exists $cmdline_arg{noctf}) {
	print AUTO "auto noctf           1\n";
    }
    print AUTO "\n";

    my $in_rad  = sprintf "%0.0f", $boxrad/3.0;
    my $out_rad = sprintf "%0.0f", $boxrad - 10.0;
    print AUTO "pcut in_rad          $in_rad\n";
    print AUTO "pcut out_rad         $out_rad\n";
    print AUTO "\n";

    my $res_min = 29.5;
    my $res_max = 29.0;
    print AUTO "p3dr res_min         $res_min\n";
    print AUTO "p3dr res_max         $res_max\n";
    print AUTO "\n";

    $res_min = 60.0;
    $res_max = 30.0;
    print AUTO "psf  res_min         $res_min\n";
    print AUTO "psf  res_max         $res_max\n";
    print AUTO "\n";

    my $annulus_low  = sprintf "%0.0f", $boxrad/3.0;
    my $annulus_high = sprintf "%0.0f", $boxrad - 10.0;
    my $resolution_low = $boxrad*2.0*$pixel_size/5.0;
    my $resolution_high = 30.0;
    my $pftrad_hi = $boxrad;
    print AUTO "ppft annulus_low     $annulus_low\n";
    print AUTO "ppft annulus_high    $annulus_high\n";
    print AUTO "ppft delta_theta     1 # Use only for RMC\n";
    print AUTO "ppft resolution_low  $resolution_low\n";
    print AUTO "ppft resolution_high $resolution_high\n";
    print AUTO "ppft pftrad_hi       $pftrad_hi\n";
    print AUTO "ppft verbose         -1 # Use only for RMC\n";

    if($pixel_size <= 4.0) {
	print AUTO "ppft bin_factor      2 # Use only for RMC\n";
    }

    print AUTO "\n";

    foreach my $pif (@pifs) {
       print AUTO "data $ptlefile{$pif}\n";
    }
    close(AUTO);

    # Create directory for random model calculation then write out
    # particle parameter files containing random orientations. If
    # relative path name was provided to images, then prefix with
    # "../" since we're going one directory down

    my $rmcdir = "${rundir}/RMC${i}";
    mkdir $rmcdir;
    chdir($rmcdir);
    foreach my $pif (@pifs) {
	my $image_file = $pif;
	unless ($image_file =~ /^\//) {
	    $image_file = "../${image_file}";
	}
	make_random_file($image_file, $ptlefile{$pif}, $head{$pif}, 
			 $numimages{$pif}, $boxrad);
    }
    chdir($startdir);
}
close(LOGLIST);

# Determine best random model
print CMD "findbest $logfile_list $rundir\n\n";

# Generate auto3dem parameter file to continue reconstruction
my $auto_master = "${startdir}_master";
$auto_master =~ s/^.*\///;
open(AUTO, ">$auto_master");

print AUTO << "EOF";
# This auto3dem input file was generated automatically by
# setup_rmc. Reasonable defaults have been set, but you may
# want to manually edit this file to override these defaults

EOF

print AUTO "auto    mode        search\n";
print AUTO "auto    niter       5\n";
print AUTO "auto    start_map   rmc.pif\n";
print AUTO "auto    rundir      $rundir\n";
if(exists $cmdline_arg{noctf}) {
    print AUTO "auto    noctf       1\n";
}
print AUTO "p3dr    res_min     25.0\n";

print AUTO << "EOF";

# If the pixel size is sufficiently small, can get very significant
# speedups by setting the following parameters

# ppft bin_factor  2  # (Use 2x2 binning of data in search mode)
# auto bin_reduce  1  # (Automatically switch to from 2x2 to unbinned)

# If the inner and outer radii of the ordered region are known,
# setting the following parameters ofter results in both better
# performance and higher quality reconstructions

# auto freeze_annulus 1  # (keep following parameters fixed)
# ppft annulus_low    n  # (n = inner radius w/ protrusions)
# ppft annulus_high   n  # (n = outer radius w/ protrusions)
# pcut in_rad         n  # (n = inner radius w/o protrusions)
# pcut out_rad        n  # (n = outer radius w/o protrusions)

# If all ptles are used in reconstructions while in search mode AND
# auto freeze_annulus is set to 1, then full output from PPFT is not
# necessary. Setting following parameter to -1 avoids calculation of
# unnecessary correlation coefficients and results in significant
# speedups.

# ppft verbose        -1

# ----------------------------------------------------------

# List of data files given below based on input to setup_rmc.
# It is your responsibility to make sure list is complete.

EOF

foreach my $ptlefile (@list) {
    print AUTO "data    $ptlefile\n";
}
close(AUTO);

	    
# Move temporary files to new directory
print CMD "remove_useless.pl $rundir $nmodels\n";
print CMD "rm -rf RMC_temp\n";
print CMD "mkdir RMC_temp\n";
print CMD "mv RMC*_master RMC_temp\n";
print CMD "mv RMC*_RESTARTS RMC_temp\n";
print CMD "mv RMC*_summary RMC_temp\n";
print CMD "mv RMC*_log RMC_temp\n";
print CMD "mv RMC*_continue RMC_temp\n";
print CMD "mv RMC_logfile_list RMC_temp\n";
print CMD "mv RMC_run RMC_temp\n";
print CMD "mv $rundir/RMC* RMC_temp\n";
close(CMD);

# Generate cleanup command
open(CLEAN, ">RMC_cleanup");
print CLEAN "rm -rf RMC_temp RMC_cleanup\n";
close(CLEAN);

# Make run and cleanup scripts executable
chmod 0744, "RMC_cleanup", "RMC_run";

print "\nInstructions:\n";
print "  Run RMC_run to construct starting model\n";
print "  Run RMC_cleanup to remove temporary files ";
print "after calculations complete\n";
print "  Starting model will be found at ${rundir}/rmc.pif\n";
print "  auto3dem input file $auto_master automatically generated\n";
print "\n";
print "--- setup_rmc done ---\n";
# --------------------------------------------------------------

sub make_random_file {

    # Purpose: Create a parameter file corresponding to a PIF
    # file. Particle orientations are chosen randomly within the
    # asymmetric unit and coordinates of origin are set equal to the
    # box radius.
    #
    # Input arguments:
    #   $pif - name of PIF file
    #   $ptlefile - name of particle parameter file
    #   $head - header line (voltage, defocus, etc.)
    #   $numimages - number of images in PIF file
    #   $boxrad - box radius

    my ($pif, $ptlefile, $head, $numimages, $boxrad) = @_;

    open(PTLE, ">$ptlefile");
    print PTLE "$pif\n";
    print PTLE "$head\n";

    for (my $i=1; $i<=$numimages; $i++) {

	# Choose random theta, |phi| in ASU
	my $j = int(rand($nviews));
	my $theta = $theta[$j];
	my $phi = $phi[$j];
    
	# Randomly assign sign to phi
	if(int(rand(2)) == 1) {
	    $phi = -$phi;
	}

	# Random select omega in range 0 to 360
	my $omega = rand(360.0);
	$omega = sprintf "%0.2f", $omega;

	printf PTLE "%5d, %7.2f, %7.2f, %7.2f, " ,
	$i, $theta, $phi, $omega;

	printf PTLE "%6.2f, %6.2f, ",
	$boxrad, $boxrad;

	printf PTLE "%5.3f, %5.3f, %5.3f, %5.3f\n",
	1.0, 0.0, 0.0, 0.0;
    }
    close(PTLE);
}

sub get_all_views {

    # Purpose: Generate a set of orientations (theta, phi) that evenly
    # cover the icosahedral asymmetric unit. For each value of theta,
    # values of phi are chosen that span the interval
    # 0<=phi<=phimax. In order to evenly sample points on a sphere, a
    # spacing $dang/sin(theta) is used. The maximum value of phi is
    # calculated by a call to subroutine get_phimax()
    #
    # Input parameters: 
    #   $dang - angular spacing for theta
    #   $theta_ref - reference to array of theta's (changed on output) 
    #   $phi_ref - reference to array of phi's (changed on output)
    #
    # Return values:
    #   None

    my ($dang, $theta_ref, $phi_ref) = @_;

    # The first orientation is always chosen to be (69.09, 0)
    my $theta = 69.09;
    my $phi = 0.0;

    push(@{$theta_ref}, $theta);
    push(@{$phi_ref}, $phi);

    # Subsequent orientations start at 69.0 + $dang
    $theta = 69.0 + $dang;

    while ($theta <= 90.0) {
	my $dphi = $dang/sin($theta/180.0*acos(-1.0));
	my $phimax = get_phimax($theta);
	$phi = 0.0;
	while ($phi <= $phimax) {
	    my $theta_rnd = sprintf "%0.2f", $theta;
	    my $phi_rnd   = sprintf "%0.2f", $phi;
	    push(@{$theta_ref}, $theta_rnd);
	    push(@{$phi_ref}, $phi_rnd);
	    $phi = $phi + $dphi;	
	}		
	$theta = $theta + $dang;
    } 
}

sub get_phimax {

    # Purpose: Calculate the maximum allowable value of the angle phi
    # in the icosahedral asymmetic unit for a given value of
    # theta. Calculation is based on spherical triognometry relations
    # described in CRC Standard Mathematical Tables
    # Input parameter:
    #   $theta - inclination angle
    #
    # Return value:
    #   $phimax - maximum allowable azimuthal angle

    my ($theta) = @_;

    my $small_b = 69.09484255211070;
    my $c       = 31.71747441146101;
    my $deg_to_rad = acos(-1.0)/180.0;
    my $rad_to_deg = 1.0/$deg_to_rad;
    my $ninety = 90.0;

    my $temp = (1/tan($c*$deg_to_rad/2))  * 
	(sin(($ninety-$small_b)*$deg_to_rad/2)/ 
	 sin(($ninety+$small_b)*$deg_to_rad/2)) ;
    my $a_minus_b = 2*$rad_to_deg*atan($temp);
    $temp = (1/tan($c*$deg_to_rad/2))  * 
	(cos(($ninety-$small_b)*$deg_to_rad/2)/ 
	 cos(($ninety+$small_b)*$deg_to_rad/2)) ;
    my $a_plus_b = 2*$rad_to_deg*atan($temp) ;
    my $b = ($a_plus_b - $a_minus_b) / 2;
    my $a = $a_minus_b + $b;
    
    $temp    = sin($a*$deg_to_rad)*sin($small_b*$deg_to_rad) / 
	sin($theta*$deg_to_rad) ;
    $b       = $rad_to_deg*asin($temp);
    $temp    = sin(($theta-$small_b)*$deg_to_rad/2)  / 
	( sin(($theta+$small_b)*$deg_to_rad/2) *  
          tan(($a-$b)*$deg_to_rad/2) );
    my $phimax  = 2*$rad_to_deg*atan($temp) ;

    return($phimax);
}
